"""
Module: laghuvu_guruvu.py
Description: Contains Telugu aksharam tokenizer, and Laghuvu-Guruvu generator functions.
Author: Boddu Sri Pavan
License: MIT
"""

import regex as re
from collections import Counter

from .nidhi import lg_map, varnamala, hallulu

class LaghuvuGuruvu:
    """
    ## Contains Aksharam Tokenizer, Laghuvu-Guruvu Generator.

    ## Attributes
    -------------
    1. data: str
        - Input Telugu string to be processed.

    ## Methods
    -----------
    1. tokenize
        - Aksharam Tokenizer.
    2. generate
        - Laghuvu-Guruvu Generator.
    """
    
    def __init__(self, data):

        # Remove leading, trailing spaces, and some unwanted characters
        self.data= data.strip().replace('\u200c', "").replace('\u200d', "").replace("x", "") 


    def tokenize( self ):
        """
        ## Aksharam Tokenizer
        ### This function tokenizes the input Telugu string into 'Aksharam Tokens'.

        ## Attributes
        -------------
        1. text: list
            - Each element corresponds to one Akharam Token for generating Laghuvu-Guruvu.

        ## Returns
        ----------
        1. text: list
            - Each element corresponds to one Akharam Token for generating Laghuvu-Guruvu.
        """
        
        # Using regex to split the raw text
        temp_l= re.findall(r"\X", self.data)

        # Remove unwanted characters and symbols
        l= []
        for i in temp_l:
            if (not i in list("""` ~ ! @ # $ % ^ & * ( ) _ - + = { } [ ] \ | ; : ' " “ ” ‘ ’ , < > . / ? ఽ ।""")) and  (not i.upper() in "ABCDEFGHIJKLMNOPQRSTUVWXYZ") and (not i.isnumeric() ):
                l.append(i)

        index= 0
        text= []
        temp= ""

        for index in range(len(l) ):

            # Ignore "Ara Sunna" (Half Circle, often pronounced similar to 'Pollu Hallu')
            l[index]= l[index].strip("ఁ")

            # Pass to next character if current character is space
            if l[index].isspace(): #or l[index].isnumeric() or l[index].upper() in "ABCDEFGHIJKLMNOPQRSTUVWXYZ" or l[index] in list("""` ~ ! @ # $ % ^ & * ( ) _ - + = { } [ ] \ | ; : ' " “ ” ‘ ’ , < > . / ? ఽ ।"""):
                # text.append( l[index] )
                pass

            # If current character is 'Pollu Hallu', 'temp' flag is empty, 'index' is less than second from last, and not last character is space
            # To capture 'Samlistaaksharam' (Conjuct) using 'temp' variable
            elif l[index].endswith('్') and temp == "" and index< len(l)-1 and not l[index+1].isspace():
                temp+= l[index]
        
            # If current character ends with 'Pollu Hallu', 'temp' is not empty and not last character is space
            # For 'Samlistaaksharam' (Conjuct) using 'temp' variable
            elif l[index].endswith('్') and temp != "" and not l[index+1].isspace():
                temp+= l[index]

            # For pollu hallu (at the end of word or at the end of the line/ sentence)
            elif l[index].endswith('్') and (index+1 == len(l) or l[index+1].isspace() ):
                text[-1]+= l[index]

            # If current character is not 'Pollu Hallu' and 'temp' flag is not empty
            # Indicates completition of 'Samlistaaksharam' (Conjuct)
            elif (not l[index].endswith('్')) and temp != "":
                text.append( temp+l[index] )
                temp= ""
            
            # If current character is not "Pollu Hallu" and 'temp' flag is empty
            elif (not l[index].endswith('్')) and temp == "":
                text.append( l[index] )
                temp= ""

            # Unknown case to handle (for future purpose)
            else:
                print("Unknown Case (for future purpose) !\n We welcome your valuable contributions to 'chandassu' !")

        self.text= text

        return self.text

    def generate( self ):
        """
        ## LaghuvuGuruvu Generator
        ### Generates Laghuvu-Guruvu for the tokenized Telugu string.

        ## Returns
        ----------
        Return a list of tuples
        where each tuple is in the format: (Aksharam Token, Laghuvu/ Guruvu)
        where 
        Aksharam Token: str
        Laghuvu/ Guruvu: str
        """

        l= self.tokenize()

        # Stores the corresponding Laghuvu-Guruvu of each generated token
        marking= []
        
        # Edge: 
        # l= ["గ","ర్భం"] #['చ', "ష్ణున్"] #['చ', 'క్రాన్'] #['వి', 'ష్ణున్']
        # re.findall(r"\X","విర్"), re.findall(r"\X", "ప్చర్"), re.findall( r"\X", "ప్చార్")

        # Iterate through the list of tokens
        for index in range( len(l) ):

            # If index upto last but one
            if index < len(l)-1 :
                
                # Split the succeeding Aksharam Token using regex
                x= re.findall(r"\X", l[index+1])

                # Check if the succeeding Aksharam Token ends with 'Pollu Hallu'
                if x[-1].endswith('్'):
                    x= "".join(x[:-1]) 
                
                else:
                    x= "".join(x)   # ["గ","ర్భం"]

                # Count the no.of occurrences of each character in the succeeding Aksharam Token
                d= Counter(x)

                # Ignore 'ర' as sometimes it is not considered in Samslistaaksharam for Guruvu Symbol
                del d["ర"]
                
                temp_count= 0
                for i in d:
                    if i in hallulu:

                        # Increment the count iff current character is "Hallu"
                        temp_count+= 1*d[i] # "పుత్త్రు"

                # 'ర' in succeeding Aksharam Token and 
                # either 'temp_count' is '0' or 'temp_count' is '1' and corresponding Aksharam Token does not start with 'ర'
                if 'ర' in l[index+1] and ((temp_count==0) or (temp_count == 1  and (not l[index+1].startswith('ర')))):
                    marking.append( lg_map[l[index][-1]] )
            
                else:

                    # Find count of varnamala characters in succeeding Aksharam Token
                    count= 0
                    for j in list(l[index+1]):
                        if j in varnamala:
                            count+= 1

                    # If 'count' is greater than '1' and succeeding Aksharam Token does not end with 'Pollu'
                    # then mark current Aksharam Token as 'U' (Guruvu)
                    if count > 1 and not l[index+1].endswith('్'):
                        marking.append( "U" )
                    
                    # If 'count' is greater than '1' and succeeding Aksharam Token ends with 'Pollu'
                    # then mark current Aksharam Token as 'U' (Guruvu)
                    elif count > 1 and l[index+1].endswith('్') and re.findall( r"\X", l[index+1])[0].endswith('్'):
                        marking.append( "U" )
                    
                    # If 'count' is greater than '1' and succeeding Aksharam Token ends with 'Pollu'
                    # the,ṁn mark current Aksharam Token as per the last character in current Aksharam Token
                    elif count > 1 and l[index+1].endswith('్'):
                        marking.append( lg_map[l[index][-1]] )

                    # Mark current Aksharam Token as per the last character in current Aksharam Token
                    else:
                        marking.append( lg_map[l[index][-1]] )
            
            # Check the rightmost character in each Aksharam Token
            elif l[index][-1] in lg_map:
                marking.append( lg_map[l[index][-1]] )

            # Unknown case to handle (for future purpose)
            else:
                print("Unknown Case (for future purpose) !\n We welcome your valuable contributions to 'chandassu' !")

        # Return a list of tuples where each tuple is in the format: (Aksharam Token, Laghuvu/ Guruvu)
        # Not dict because dict donot allow multiple keys with same name
        return list(zip(l, marking))
    
def tokenize( data ):
    """
    ## Aksharam Tokenizer
    ### This function tokenizes the input Telugu string into 'Aksharam Tokens'.

    ## Attributes
    -------------
    1. text: list
        - Each element corresponds to one Akharam Token for generating Laghuvu-Guruvu.

    ## Returns
    ----------
    1. text: list
        - Each element corresponds to one Akharam Token for generating Laghuvu-Guruvu.
    """

    return LaghuvuGuruvu(data= data).tokenize()

def generate( data ):
    """
    ## LaghuvuGuruvu Generator
    ### Generates Laghuvu-Guruvu for the tokenized Telugu string.

    ## Returns
    ----------
    Return a list of tuples
    where each tuple is in the format: (Aksharam Token, Laghuvu/ Guruvu)
    where 
    Aksharam Token: str
    Laghuvu/ Guruvu: str
    """

    return LaghuvuGuruvu(data= data).generate()


"""
Module: panimuttu.py
Description: Contains utility functions
Author: Boddu Sri Pavan
License: MIT
"""

from .nidhi import gunintha_chihnam, varnamala

def remove_gunintha_chihnam( x ):
    """Remove all the Gunintha Chihnams from the input text."""
    
    l= ""
    for i in x:
        if i not in gunintha_chihnam:
            l+= i
    return l

def extract_gunintha_chihnam( x ):
    """Extract Gunintha Chihnam from the given Aksharam Token."""
    
    l= ""
    for i in gunintha_chihnam:
        if i in x:
            l= i
    
    # For vowel sound: 'అ'
    if l== "":
        l= " "
        
    return l

def extract_aksharam( x ):
    """Extracts all varnamala letters into a list."""
    
    l= []
    for i in x:
        if i in varnamala:
            l.append(i)
    return l

def extract_paadam( padyam ):
    """Extracts paadams (lines) from given padyam based on newline separator (for future purpose)."""
    
    l= []
    for i in padyam.split("\n"):
        
        i= i.strip()
        if len(i) > 0:      # Assume atleast one letter present in the padyam
            l.append(i) 

    return l


"""
Module: lakshanam.py
Description: Contains functions to check lakshanamulu (constraints/ features) of padyam.
Author: Boddu Sri Pavan
License: MIT
"""

from .laghuvu_guruvu import LaghuvuGuruvu
from .nidhi import achhulu, yati, hraswa_chihnam, deergha_chihnam
from .panimuttu import *
from .ganam import *

import math

def check_yati( yati_sthanam= None, paadam= None, first_letter= None, yati_sthanam_letter= None, verbose= True ):
    """
    ## Check whether Paadam (Padyam line) is satisfying Yati constraint.
    
    ## Parameters
    -------------
    1. yati_sthanam: int
        - Position of corresponding Yati (Natural Number as per Telugu literature).
    2. paadam: list
        - LaghuvuGuruvu data.
    3. first_letter: str
        - Aksharam Token at Yati position (Generally the first Aksharam Token in Padyam).
    4. yati_sthanam_letter: str
        - Corresponding position to Yati position as per Padyam configuration.
    5. verbose: bool
        - Prints the result of each step.
        - For traceability.
        - Default is set to 'False'.

    ## Returns
    ----------
    'True' is Yati constraint is satisfied else 'False'.
    """
    
    # Initialize 'first_letter' and 'yati_sthanam_letter' if not passed to the function
    if first_letter == None and yati_sthanam_letter == None :
        if verbose:
            print( paadam )

        first_letter= paadam[0]
        yati_sthanam_letter= paadam[ yati_sthanam-1 ]

    # Ignoring Visarga and PurnaBindu (As not mentioned in the reference book)
    first_letter= first_letter.replace('ం', "").replace('ః', "")
    first_letter= first_letter.replace("ಂ", "").replace( 'ః', "")
    yati_sthanam_letter= yati_sthanam_letter.replace('ం', "").replace('ః', "")
    yati_sthanam_letter= yati_sthanam_letter.replace("ಂ", "").replace('ః', "")

    if verbose:
        print("First aksharam (letter): ", first_letter)
        print("Yati sthana aksharam (letter): ", yati_sthanam_letter)

    # Check Samslistaaksharam (conjuct)
    samdit= False
    if len(extract_aksharam(first_letter))> 1:
        samdit= True

    if samdit or (not samdit):
        
        # Extract Gunintha Chihnam
        chihnam_a= extract_gunintha_chihnam( first_letter )
        chihnam_b= extract_gunintha_chihnam( yati_sthanam_letter )

        # Chihna Yati: Yati checking for vowel symbol
        chihna_yati= False
        for i in yati:

            if chihnam_a in i:

                if chihnam_b in i:
                    chihna_yati= True

                else:
                    if verbose:
                        print(f"Chihna yati mismatch occurred between '{chihnam_a}' in '{first_letter}'  and '{chihnam_b}' in '{yati_sthanam_letter}'")
                    return False

        # Akshara Yati: Yati check for other than vowel sound
        # సంయుక్తాక్షరాలు వచ్చిన చోట, యతి కోసం ఏ అక్షరాన్నైనా గణించవచ్చు. ఉదా: "క్రొ" మొదటి అక్షరం అనుకోండి. యతి మైత్రి కోసం దీన్ని "కొ"గా గానీ "రొ"గా గానీ భావించ వచ్చు. 
        akshara_yati= False
        for i in list(set(extract_aksharam(first_letter))):

            for j in yati:
                
                if i in j:
                    
                    for k in list(set(extract_aksharam(yati_sthanam_letter))):
                        
                        if k in j:

                            akshara_yati= True

                            if verbose:
                                print(i, j, k)

                            # Atleast one of samyukta dwitwa aksharam is enough for yati
                            break
                    
                    if akshara_yati == True:
                        break
        
        if not akshara_yati:

            if verbose:
                print(f"Yati mismatch occurred between '{i}' in '{first_letter}' and '{yati_sthanam_letter}'")

            return False


        if chihna_yati and akshara_yati:

            if verbose:
                print( "Yati Matched !")

            return True

        else:

            if verbose:
                print("Chihna Yati: ", chihna_yati)
                print("Akshara Yati: ", akshara_yati)

            return False
        
    # else:

    #     if len(first_letter)==1 and first_letter not in achhulu:
    #         first_letter= [first_letter]+[' ']

    #     if len(yati_sthanam_letter)==1 and first_letter not in achhulu:
    #         yati_sthanam_letter= [yati_sthanam_letter]+[" "]

    #     to_return= False
    #     temp= []
    #     for i in first_letter:

    #         for j in yati:

    #             if i in j:
                    
    #                 flag= False
    #                 for k in yati_sthanam_letter:
    #                     if k in j:
    #                         flag= True
    #                         break

    #                 if flag== True:
    #                     temp.append( True )
    #                 else:
    #                     temp.append( False )
    #                     # No break statement should be used here
    #                     # Because same aksharam could be present in multiple associations

    #                     if verbose:
    #                         print(f"Yati mismatch occurred between: '{i}' in '{first_letter}' and '{yati_sthanam_letter}'")
        
    #     if all( temp ):
    #         if verbose:
    #             print( "Yati Matched Successfully!" )
    #         to_return= True

        # return to_return

def check_prasa_yati( padamwise_ganam_data, type, config, only_generic_yati= False, verbose= True):
    """
    ## Check Prasa Yati for given Paadamwise Laghuvu-Guruvu data.
    
    ## Parameters
    -------------
    1. padamwise_ganam_data: list
        - List of Ganamwise LaghuvuGuruvu data with Matched/UnMatched flag.
    2. type: str
        - Type of Padyam.
    3. config: dict
        - Configuration of Padyam.
    4. only_generic_yati: bool
        - Flag to apply Prasa Yati.
        - Default is set to 'False'.
    5. verbose: bool
        - Prints the result of each step.
        - For traceability.
        - Default is set to 'False'.

    ## Returns
    ----------
    List of boolean values where index(i) corresponds to (i+1)th Paadam (Padyam Line) Yati match value.
    """
    
    if verbose:
        print( "Paadalu to follow Yati: ", config["yati_paadalu"] )
        print( "No.of paadalu to follow Yati: ", len(config["yati_paadalu"]))

    # Kanda Padyam Yati: 2,4 Paadalu only
    # Remaining Padyams all paadams are having Yati
    if type == "kandamu":
        padamwise_ganam_data= [ padamwise_ganam_data[i-1] for i in config["yati_paadalu"] ] 
    
    if verbose:
        print("Updated paadamwise_ganam_data: ")
        print( padamwise_ganam_data )
        
    yati_match= []

    # Iterate through each Paadam (Padyam line)
    for row in padamwise_ganam_data:
        
        if verbose:
            print(row)

        if (len(row[0][0]) > 1) and (len(row[ config["yati_sthanam"][0]-1 ][0]) > 1 ):

            # Extract Aksharam Token in Yati position
            first_letter= [a[0] for a in row[0][0]]

            # Extract Aksharam Token in corresponding position to Yati as per Padyam configuration
            yati_sthanam_letter= row[ config["yati_sthanam"][0]-1 ][0][ config["yati_sthanam"][1] ]
            
            # Check 'Generic Yati' (Not Prasa Yati)
            generic_yati= check_yati( first_letter= first_letter[0], yati_sthanam_letter= yati_sthanam_letter[0], verbose= verbose )
            
            # If 'Generic Yati' constraint is satisfied then no need to check 'Prasa Yati' (as per current Padyam types)
            if only_generic_yati:

                if verbose:
                    print( "Generic Yati: ", generic_yati )
                    print( first_letter, yati_sthanam_letter )

                yati_match.append( generic_yati )

                continue
            
            # This extracts two aksharas yati-sthanam letter and its succeeding (For Prasa Yati implementation)
            yati_sthanam_letter= [a[0] for a in row[ config["yati_sthanam"][0]-1 ][0]][:2]

            prasa_yati_match= False

            if generic_yati:
                if verbose:
                    print("Generic Yati Matched")
                yati_match.append( True )

            else:
                
                # Reference: https://te.wikipedia.org/wiki/%E0%B0%AA%E0%B1%8D%E0%B0%B0%E0%B0%BE%E0%B0%B8%E0%B0%AF%E0%B0%A4%E0%B0%BF

                # Yati Sthanam: Hraswa-Deergham Check
                hraswa_deergham_flag_1= " "

                if first_letter[0][-1] in gunintha_chihnam:
                    hraswa_deergham_flag_1= first_letter[0][-1]
                else:
                    hraswa_deergham_flag_1= " "

                hraswa_deergham_flag_2= ""

                if yati_sthanam_letter[0][-1] in gunintha_chihnam:
                    hraswa_deergham_flag_2= yati_sthanam_letter[0][-1]
                else:
                    hraswa_deergham_flag_2= " "


                # Prasa Sthanam: Prasa Check
                l1= remove_gunintha_chihnam( first_letter[1] )
                l2= remove_gunintha_chihnam( yati_sthanam_letter[1] )
                
                # Yati Sthanam
                if ( hraswa_deergham_flag_1 in hraswa_chihnam and hraswa_deergham_flag_2 in hraswa_chihnam ):
                    
                    if verbose:
                        print("Both Yati Sthanam aksharam: Hraswa")

                    # Prasa Sthanam
                    if  l1 == l2 :
                        prasa_yati_match= True
                        if verbose:
                            print("Both Yati Sthanam chihnam matched")
                    else:
                        if verbose:
                            print("Second Letter Mismatched: ", l1, "===", l2)

                elif ( hraswa_deergham_flag_1 in deergha_chihnam and hraswa_deergham_flag_2 in deergha_chihnam ):
                    
                    if verbose:
                        print("Both Yati Sthanam aksharam: Deergham ")

                    if l1 == l2:
                        prasa_yati_match= True
                        if verbose:
                            print("Both Yati Sthanam chihnam matched")
                    else:
                        if verbose:
                            print("Second Letter Mismatched: ", l1, "===", l2)

                else:
                    if verbose:
                        print("First Akshara Chihnam in Prasa Yati Mis-matched")
                        print("First Akshara Chihnam: ", hraswa_deergham_flag_1)
                        print("Prasa Yathi Akshara Chihnam: ", hraswa_deergham_flag_2)

                if prasa_yati_match:
                    yati_match.append( True )
                    
                    if verbose:
                        print("Prasa Yati Matched")

                else:

                    if verbose:
                        print("Prasa Yati Mis-matched")

            # If 'generic_yati' and  'prasa_yati_match' did not match
            if generic_yati == False and prasa_yati_match == False:
                yati_match.append( False )
                if verbose:
                    print("Both Yati and PrasaYati mis-matched")
        else:
            yati_match.append( False )
            
            if verbose:
                print("No paadam found")

    return yati_match





def n_aksharam( data, verbose= True ):
    """
    ## Counts no.of Aksharam Tokens from the given LaghuvuGuruvu data.

    ## Parameters
    -------------
    1. data: list
        - LaghuvuGuruvu data.
    2. verbose: bool
        - Prints the result of each step.
        - For traceability.
        - Default is set to 'False'.
    
    ## Note
    -------
    For future purpose.
    """
    
    # Implements same functionality
    # n_letters= []
    # for i in p.split("\n"):
    #     lg= LaghuvuGuruvu( data= i.strip() )
    #     n_letters.append( len(lg.tokenize()) )
    # return n_letters

    n= len(data)

    if verbose:
        print("No.of aksharams (letters): ", n)
        print()

    return n

def n_paadam( data, aksharam_per_paadam, clip= False, verbose= True ):
    """
    ## Count no.of paadams (lines) in given padyam.
    
    ## Parameters
    -------------
    1. data: list
        - LaghuvuGuruvu data.
    2. aksharam_per_paadam: int
        - No.of Aksharam Tokens for Paadam (Padyam Line).
    3. clip: bool
        - To apply floor.
    4. verbose: str
        - Flag to print tracing steps.
    
    ## Returns
    ----------
    Count (int) of no.of paadams.

    ## Note
    -------
    For future purpose.
    """

    n= len(data)/ aksharam_per_paadam

    if clip:
        if n-int(n) != 0:
            n= math.floor( n )
    if verbose:
        print("No.of paadams (lines) found: ", n)
        print()
        
    return n


"""
Module: padya_bhedam.py
Description: Contains functions to check type of Padyam.
Author: Boddu Sri Pavan
License: MIT
"""

from .check_lakshanam import *
from .padyam_config import *

TYPE_TO_BHEDAM_MAP= {
                        'kandamu': Jaathi,

                        'aataveladi': VupaJaathi, 
                        'teytageethi': VupaJaathi, 
                        'seesamu': VupaJaathi,

                        'vutpalamaala': Vruttamu, 
                        'champakamaala': Vruttamu,
                        'mattebhamu': Vruttamu,
                        'saardulamu': Vruttamu
                    }

VRUTTAMU= ["vutpalamaala", "champakamaala", "mattebhamu", "saardulamu"]

def check_padyam( lg_data, type= "aataveladi", return_micro_score= True, verbose= False):
    """
    ## Evaluates given Laghuvu-Guruvu data with given padyam type with confidence scores.
    
    ## Parameters
    -------------
    1. lg_data: list
        - Laghuvu-Guruvu data generated using laghuvu_guruvu.LaghuvuGuruvu.generate()
    2. type: str
        - Type of padyam
        - Supported types: 'kandamu', 'aataveladi', 'teytageethi', 'seesamu', 'vutpalamaala',
                            'champakamaala', 'mattebhamu', 'saardulamu'
    3. return_micro_score: bool
        - Set to 'True' to return lakshanamwise scores (micro scores).
        - Default is set to 'True'.
    4. verbose: bool
        - Prints the result of each step.
        - For traceability.
        - Default is set to 'False'.

    ## Returns
    ----------
    Dictionary of scores (Chandassu Score and Micro Score).
    """
    
    bhedam= TYPE_TO_BHEDAM_MAP[type]

    if verbose:
        print("Type: ", type)
        print("Bhedam: ", bhedam)

    config= getattr(bhedam, type, False)
    
    # Holds Laghuvu-Guruvu data for each Paadam (Padyam line)
    # Data Format
    # -----------
    # padamwise_ganam_data= [ paadam_1, paadam_2, paadam_3, ..., paadam_n ]
    # paadam_n= [ ganam_match_1, ganam_match_2, ganam_match_3, ..., ganam_match_n ]
    # ganam_match_n= [ LaghuvuGuruvu_data, Matched/ UnMatched flag]
    padamwise_ganam_data= []

    gana_kramam_score= 0
    end= 0
    paadam_count= 0

    # Parse with respect to each line of padyam (Paadam)
    for line in range( len( config["gana_kramam"] ) ):

        ganam_data= []

        for j in range( len(config["gana_kramam"][line]) ):

            ganam_match_flag= False
            
            if verbose:
                print( config["gana_kramam"][line][j])

            for i in config["gana_kramam"][line][j]:

                # Take legth of corresponding ganam
                ganam= tuple([k[1] for k in lg_data[end: end+len(ganamulu[i]) ]])
                
                if verbose:
                    print("Ganam : ", ganam)

                try:
                    # If matched
                    if r_ganamulu[ ganam ] == i:
                        
                        ganam_data.append( [lg_data[end: end+len(ganamulu[i])], r_ganamulu[ganam]] )

                        # Increment 'Gana Kramam Score'
                        gana_kramam_score+= 1
                        
                        if verbose:
                            print( [lg_data[end: end+len(ganamulu[i])], r_ganamulu[ganam]] )

                        ganam_match_flag= True

                        break
                
                except KeyError as e:

                    if verbose:
                        print( "Key Not Found: ", ganam )
                        print( "Exception: ", str(e) )

                except Exception as e:
                    print(e)
                    pass
            
            # If ganam not matched then unmatched tracing
            if ganam_match_flag == False:
                ganam_data.append( [lg_data[end: end+len(ganamulu[i])], "UnMatched"] )
            
            # Increment with the last ganam length (maximum)
            end+= len(ganamulu[i])

        if verbose:
            print( line, end, len(lg_data), ganam_data )
        
        # Consider atleast one character to consider new line as paadam
        if len(ganam_data[0][0]) > 1:
            paadam_count+= 1

        padamwise_ganam_data.append( ganam_data )
        
        if end == len(lg_data):
            if verbose:
                print("Paadam wise split completed")
            break
    
    # Check Yati/ Prasa Yati as per Padyam configuration
    prasa_yati_match= check_prasa_yati( padamwise_ganam_data= padamwise_ganam_data, type= type, config= config, verbose= verbose, only_generic_yati= config["only_generic_yati"])

    if verbose:
        print("Paadam Count: ", paadam_count)
        print("Ganam Kramam Score: ", gana_kramam_score)
        print("Paadam-wise Yati: ", prasa_yati_match)


    N_PAADALU= config.get("true_n_paadalu", -1)
    if N_PAADALU == -1: # Not Present
        N_PAADALU= config["n_paadalu"]

        if verbose:
            print("true_n_paadalu parameter found in config", N_PAADALU)

    total_yati_paadalu= len(config["yati_paadalu"])

    # Seesamu padyam, for convineience considered pseudo paadam count as 8
    # Therefore, halving it
    if type== "seesamu":
        paadam_count= paadam_count/2
    
    # Calculate lakshanamwise scores (micro scores)
    score= {
                'n_paadalu':  paadam_count/ N_PAADALU,
                'gana_kramam': gana_kramam_score/ sum([len(i) for i in config["gana_kramam"]]),
                'yati_sthanam': sum(prasa_yati_match)/ total_yati_paadalu
            }

    # This conditions becomes True only for Vruttamu
    if config.get("n_aksharalu", False):

        # Aksharam Count Score
        aksharam_count= 0

        # Iterate through each Paadam (Padyam Line)
        for i in padamwise_ganam_data:
            
            if verbose:
                print(i)

            for j in i:

                if verbose:   
                    print(j[0], len(j[0]))

                aksharam_count+= len(j[0])

        # This is to quantify the overflow as unexpected
        # This will activate only when No.of aksharams in given > No.of aksharams in expected
        # Else: 0
        subtract_factor= len(lg_data) - aksharam_count

        score["n_aksharalu"]= (aksharam_count-subtract_factor)/ (N_PAADALU*config["n_aksharalu"])

    # This conditions becomes True only for Vruttamu and Jaathi
    if config.get("prasa", False):

        # Prasa Score
        index= 2    # Second letter

        frequency= {}

        # Iterate through each Paadam (Padyam Line)
        for i in padamwise_ganam_data:

            try:
                aksharam= remove_gunintha_chihnam(i[0][0][ index-1 ][0])
                frequency[aksharam]= frequency.get( aksharam , 0) + 1
                
                if verbose:
                    print(aksharam)
            except:
                pass

        if verbose:
            print( "Frequency of second aksharam (letter): ", frequency )
            if len( frequency ) != 1:
                print("Prasa Mismatch Occurred : ", frequency )
                print()

            elif (len(frequency) == 1) and (max(frequency.values()) == config["n_paadalu"]): # As 'seesamu' has no necessary condition to check prasa, can be ignored
                print("Prasa Matched Successfully !")
                print()

            else:
                print("Prasa Mismatch Occurred : ", frequency )
                print()

        score["prasa"]= max( frequency.values() )/ N_PAADALU

    # Calculate 'Chandassu Score': Average of all lakshanamwise scores for given Padyam type
    overall_score= sum(score.values()) / len(score)

    if return_micro_score:
            return {"chandassu_score": overall_score, "micro_score": score}

    return {"chandassu_score": overall_score}




# def check_teytageethi( lg_data, type= "teytageethi", verbose= True, weightage_normalization= True,
#                         weightage_factor= {"n_paadalu": 1, "gana_kramam": 1, "yati_sthanam": 1}, return_micro_score= True
#                     ):

#     try:

#         # Weightage factor parameter check
#         for i in ["n_paadalu", "gana_kramam", "yati_sthanam"]:
#             if i not in weightage_factor:
#                 print( "Not present in 'weightage_factor': ", i )
#                 return False

#         if weightage_normalization:

#             total= sum(weightage_factor.values())

#             weightage_factor= {i:j/total for i, j in weightage_factor.items()}

#         config= getattr(VupaJaathi, type, False)

#         padamwise_ganam_data= []

#         gana_kramam_score= 0
#         end= 0
#         paadam_count= 0

#         while end< len(lg_data):

#             ganam_data= []

#             for j in range( len(config["gana_kramam"]) ):

#                 for i in config["gana_kramam"][j]:

#                     # Take legth of corresponding ganam
#                     ganam= tuple([k[1] for k in lg_data[end: end+len(ganamulu[i]) ]])

#                     if verbose:
#                         print("Ganam : ", ganam)

#                     try:
#                         if r_ganamulu[ ganam ] == i:
                            
#                             ganam_data.append( [lg_data[end: end+len(ganamulu[i])], r_ganamulu[ganam]] )

#                             gana_kramam_score+= 1
                            
#                             if verbose:
#                                 print( [lg_data[end: end+len(ganamulu[i])], r_ganamulu[ganam]] )

#                             break
                
#                     except Exception as e:
#                         print(e)
#                         pass

#                 # Increment with the last ganam length (maximum)
#                 end+= len(ganamulu[i])
            
#             paadam_count+= 1
#             padamwise_ganam_data.append( ganam_data )

#             if verbose:
#                 print( ganam_data )

#         match_yati= []
#         for line in padamwise_ganam_data:

#             if verbose:
#                 print(line)
            
#             yati_value= check_yati(     
#                                         yati_sthanam= True,
#                                         paadam= line, 
#                                         first_letter= line[0][0][0][0], 
#                                         yati_sthanam_letter= line[config["yati_sthanam"][0]-1][0][0][0], 
#                                         verbose= verbose 
#                                     )
            
#             match_yati.append(yati_value)

#             if verbose:
#                 print( yati_value )
    

#         score= {
#                     'n_paadalu': weightage_factor['n_paadalu']* paadam_count/ config["n_paadalu"],
#                     'gana_kramam': weightage_factor['gana_kramam']* gana_kramam_score/ (config["n_paadalu"]*len(config["gana_kramam"])),
#                     'yati_sthanam': weightage_factor['yati_sthanam']* sum(match_yati)/ config["n_paadalu"],
#                 }
        
#         overall_score= sum(score.values())

#         if overall_score == 1:
#             print("Padyam Detected: ", type.upper())
        
#         else:
#             print("Padyam not exactly matched with: ", type.upper())

#         for i in score:
#             score[i]= (score[i], weightage_factor[i])
        
#         if return_micro_score:
#             return {"chandassu_score": overall_score, "micro_score": score}
        
#         return {"chandassu_score": overall_score}
    
#     except Exception as e:
#         print( "Exception Occurred: ", str(e) )

# def check_vruttam( 
#                     lg_data, type, verbose= True, weightage_normalization= True,
#                     weightage_factor= {"n_paadalu": 1, "n_aksharalu": 1, "gana_kramam": 1, "yati_sthanam": 1, "prasa": 1},
#                     return_micro_score= True                  
#                 ):

#     try:

#         # Weightage factor parameter check
#         for i in ["n_paadalu", "n_aksharalu", "gana_kramam", "yati_sthanam", "prasa"]:
#             if i not in weightage_factor:
#                 print( "Not present in 'weightage_factor': ", i )
#                 return False
            
#         if weightage_normalization:

#             total= sum(weightage_factor.values())

#             weightage_factor= {i:j/total for i, j in weightage_factor.items()}
            

#         config= getattr(Vruttamu, type, False)

#         if verbose:
#             print(config)

#         # There are 5 lakshanams (constraints/ rules) for a Vrutta padyam to be satisfied.
#         # Following Macro technique as each one is given equal weightage.
#         # Therefore, multiply each one with 1/5 = 0.2
#         score= {
#                     'n_paadalu': 0,
#                     'n_aksharalu': 0,
#                     'gana_kramam': 0,
#                     'yati_sthanam': 0,
#                     'prasa': 0
#                 }
        
#         # 1. Check no.of paadams (lines)
#         count_paadam= n_paadam( data= lg_data, aksharam_per_paadam= 20, clip= True, verbose= verbose )

#         # 2. Check no.of aksharams (letters)
#         count_aksharam= n_aksharam( data= lg_data, verbose= verbose )
        
#         aksharam_per_paadam= config["n_aksharalu"]

#         paadam_lg= [ lg_data[i*aksharam_per_paadam: (i+1)*aksharam_per_paadam] for i in range(4)]


#         # 3. Match gana kramam in each paadam
#         _, gana_kramam_score= check_vruttam_gana_kramam( lg_data, config, verbose= verbose )

#         if verbose:
#             print()

#         # 4. Match yati in each paadam
#         match_yati= []
#         for i in range(len(paadam_lg)): #lines:
#             line= [j[0] for j in paadam_lg[i]]

#             try:
#                 yati_value= check_yati( paadam= line, yati_sthanam= config['yati_sthanam'], verbose= verbose )
#                 match_yati.append( yati_value )
#             except:
#                 # Condition where no.of letters in a paadam are lesser than the yat number
#                 match_yati.append( False )

#         try:
#             o= [[paadam_lg[a][b][0] for b in range(len(paadam_lg[a]))] for a in range(len(paadam_lg))]
#         except Exception as e:
#             print(">>>>>", e)

#         # 5. Match prasa in each paadam
#         match_prasa= check_prasa( padya_paadaalu= [[paadam_lg[a][b][0] for b in range(len(paadam_lg[a]))] for a in range(len(paadam_lg))], 
#                                 index= 2, 
#                                 verbose= verbose
#                                 )

#         # 'weightage_factor' can be modified for more insights
#         score["n_paadalu"]= weightage_factor["n_paadalu"]*count_paadam/ config["n_paadalu"]
#         score["n_aksharalu"]= weightage_factor["n_aksharalu"]* count_aksharam/ (config["n_paadalu"]*config["n_aksharalu"])
#         score["gana_kramam"]= weightage_factor["gana_kramam"]* gana_kramam_score/ (config["n_paadalu"]*len(config["gana_kramam"]))
#         score["yati_sthanam"]= weightage_factor["yati_sthanam"]*sum(match_yati)/ config["n_paadalu"]
#         score["prasa"]= weightage_factor["prasa"]*max( match_prasa.values() )/ config["n_paadalu"]

#         for sub_score in ["n_paadalu", "n_aksharalu", "gana_kramam", "yati_sthanam", "prasa"]:

#             # Implementing compliment to avoid overflow
#             if score[ sub_score ] > 1:
#                 score[sub_score]= 1- score[sub_score]

#         overall_score= sum(score.values())

#         if overall_score == 1:
#             print("Padyam Detected: ", type.upper())
        
#         else:
#             print("Padyam not exactly matched with: ", type.upper())

#         for i in score:
#             score[i]= (score[i], weightage_factor[i])

#         if return_micro_score:
#             return {"chandassu_score": overall_score, "micro_score": score}

#         return {"chandassu_score": overall_score}

#     except Exception as e:
#         print("==========================================")
#         print( "Given Padyam is not detected as: ", type )
#         print( "Exception Occurred: ", str(e))
 


"""
Module: padyam_config.py
Description: Contains pre-defined configurations for each type of Padyam.
Author: Boddu Sri Pavan
License: MIT
"""

from .ganam import *

class Jaathi:
    """
    ## Configurations for 'Jaathi' class of padyams.
    
    ## Attributes (Type of Padyam)
    ------------------------------
    1. kandamu: కందము
    """

    kandamu= {
                "n_paadalu": 4,
                "gana_kramam": ( ( kandam_odd, kandam_all, kandam_odd ),
                                 ( kandam_all, kandam_odd, kandam_6, kandam_odd, kandam_2_4_last ),
                                 ( kandam_odd, kandam_all, kandam_odd ),
                                 ( kandam_all, kandam_odd, kandam_6, kandam_odd, kandam_2_4_last )
                                ),
                # Here (x,y)= x is the ganam number in human notation, and y is the computer index (can be '0' zero)
                "yati_sthanam": (4, 0),   ### Logic update for configuration
                
                # Here, paadam numbers are in human notation 
                # Therefore in the Yati calculation, to get computer index, '1' will be subtracted
                "yati_paadalu": (2,4),
                "prasa": True,
                "only_generic_yati": True
            }

class VupaJaathi:
    """
    ## Configurations for 'VupaJaathi' class of padyams.
    
    ## Attributes (Type of Padyam)
    ------------------------------
    1. aataveladi: ఆటవెలది 
    2. teytageethi: తేటగీతి 
    3. seesamu: సీసము
    """
    
    aataveladi= {
                    "n_paadalu": 4,
                    "gana_kramam": ( ( surya_ganam, surya_ganam, surya_ganam, indra_ganam, indra_ganam ),
                                     ( surya_ganam, surya_ganam, surya_ganam, surya_ganam, surya_ganam ),
                                     ( surya_ganam, surya_ganam, surya_ganam, indra_ganam, indra_ganam ),
                                     ( surya_ganam, surya_ganam, surya_ganam, surya_ganam, surya_ganam )
                                    ),

                    "yati_sthanam": (4,0), # 1
                    "yati_paadalu": (1,2,3,4),
                    "only_generic_yati": False
                }

    teytageethi= {
                    "n_paadalu": 4,
                    "gana_kramam": ( ( surya_ganam, indra_ganam, indra_ganam, surya_ganam, surya_ganam ),
                                     ( surya_ganam, indra_ganam, indra_ganam, surya_ganam, surya_ganam ),
                                     ( surya_ganam, indra_ganam, indra_ganam, surya_ganam, surya_ganam ),
                                     ( surya_ganam, indra_ganam, indra_ganam, surya_ganam, surya_ganam )
                                    ),
                    "yati_sthanam": (4, 0), # 1
                    "yati_paadalu": (1,2,3,4),
                    "only_generic_yati": False
                 }
    
    seesamu= {
                "true_n_paadalu": 4,
                "n_paadalu": 8,
                "gana_kramam": (
                                    ( indra_ganam, indra_ganam, indra_ganam, indra_ganam ),
                                    ( indra_ganam, indra_ganam, surya_ganam, surya_ganam ),

                                    ( indra_ganam, indra_ganam, indra_ganam, indra_ganam ),
                                    ( indra_ganam, indra_ganam, surya_ganam, surya_ganam ),

                                    ( indra_ganam, indra_ganam, indra_ganam, indra_ganam ),
                                    ( indra_ganam, indra_ganam, surya_ganam, surya_ganam ),

                                    ( indra_ganam, indra_ganam, indra_ganam, indra_ganam ),
                                    ( indra_ganam, indra_ganam, surya_ganam, surya_ganam ),
                                ),

                "yati_sthanam": (3, 0), # 1
                "yati_paadalu": (1,2,3,4,5,6,7,8),
                "only_generic_yati": False
                # PRasa is optional
                # Therefore ignored here (for score calculation)
            }

class Vruttamu:
    """
    ## Configurations for 'Vruttamu' class of padyams.
    
    ## Attributes (Type of Padyam)
    ------------------------------
    1. vutpalamaala: ఉత్పలమాల 
    2. champakamaala: చంపకమాల
    3. saardulamu: శార్దూలము
    4. mattebhamu: మత్తేభము
    """
    
    vutpalamaala= {
                    "n_paadalu": 4,
                    "n_aksharalu": 20,
                    "gana_kramam": ( (bha_ganam, ra_ganam, na_ganam, bha_ganam, bha_ganam, ra_ganam, va_ganam),
                                     (bha_ganam, ra_ganam, na_ganam, bha_ganam, bha_ganam, ra_ganam, va_ganam),
                                     (bha_ganam, ra_ganam, na_ganam, bha_ganam, bha_ganam, ra_ganam, va_ganam),
                                     (bha_ganam, ra_ganam, na_ganam, bha_ganam, bha_ganam, ra_ganam, va_ganam),
                                    ),

                    # Here (x,y)= x is the ganam number in human notation, and y is the computer index (can be '0' zero)
                    "yati_sthanam": (4,0), # Each ganam contains 3 aksharaas. 3X3 + 1= 10 
                    "yati_paadalu": (1,2,3,4),
                    "prasa": True,
                    "only_generic_yati": True
                }

    champakamaala= {
                    "n_paadalu": 4,
                    "n_aksharalu": 21,
                    "gana_kramam": ( (na_ganam, ja_ganam, bha_ganam, ja_ganam, ja_ganam, ja_ganam, ra_ganam),
                                     (na_ganam, ja_ganam, bha_ganam, ja_ganam, ja_ganam, ja_ganam, ra_ganam),
                                     (na_ganam, ja_ganam, bha_ganam, ja_ganam, ja_ganam, ja_ganam, ra_ganam),
                                     (na_ganam, ja_ganam, bha_ganam, ja_ganam, ja_ganam, ja_ganam, ra_ganam)
                                    ),
                    
                    # Here (x,y)= x is the ganam number in human notation, and y is the computer index (can be '0' zero)
                    "yati_sthanam": (4, 1), # 11 = 3X3 + 2+1
                    "yati_paadalu": (1,2,3,4),
                    "prasa": True,
                    "only_generic_yati": True
                }


    saardulamu= {
                    "n_paadalu": 4,
                    "n_aksharalu": 19,
                    "gana_kramam": ( (ma_ganam, sa_ganam, ja_ganam, sa_ganam, ta_ganam, ta_ganam, ga_ganam),
                                     (ma_ganam, sa_ganam, ja_ganam, sa_ganam, ta_ganam, ta_ganam, ga_ganam),
                                     (ma_ganam, sa_ganam, ja_ganam, sa_ganam, ta_ganam, ta_ganam, ga_ganam),
                                     (ma_ganam, sa_ganam, ja_ganam, sa_ganam, ta_ganam, ta_ganam, ga_ganam)
                                    ),

                    # Here (x,y)= x is the ganam number in human notation, and y is the computer index (can be '0' zero)
                    "yati_sthanam": (5, 0), #13,
                    "yati_paadalu": (1,2,3,4),
                    "prasa": True,
                    "only_generic_yati": True
                }

    mattebhamu= {
                    "n_paadalu": 4,
                    "n_aksharalu": 20,
                    "gana_kramam": ( (sa_ganam, bha_ganam, ra_ganam, na_ganam, ma_ganam, ya_ganam, va_ganam),
                                     (sa_ganam, bha_ganam, ra_ganam, na_ganam, ma_ganam, ya_ganam, va_ganam),
                                     (sa_ganam, bha_ganam, ra_ganam, na_ganam, ma_ganam, ya_ganam, va_ganam),
                                     (sa_ganam, bha_ganam, ra_ganam, na_ganam, ma_ganam, ya_ganam, va_ganam)
                                    ),
                    
                    # Here (x,y)= x is the ganam number in human notation, and y is the computer index (can be '0' zero)
                    "yati_sthanam": (5, 1), #14,
                    "yati_paadalu": (1,2,3,4),
                    "prasa": True,
                    "only_generic_yati": True
                }

    # Future Implementation
    # mattakokila= {
    #                 "n_paadalu": 4,
    #                 "n_aksharalu": 18,
    #                 "gana_kramam": ('ర', 'స', 'జ', 'జ', 'భ', 'ర'),
    #                 "yati_sthanam": 11,
    #                 "prasa": True
    #             }